#!/usr/bin/perl -w
use strict; 

################################################################################
# heavily modified from the otto jukebox image downloader 
#
# this program operates in two fundamental ways : batch from the cmd line
# and called directly from the Jukebox program. 
# in cmd line mode it scans the Globecom Jukebox DB programs table looking
# for blank entries in the 'cover' column. It then does a search within either
# CDNow or Google and constructs a directory tree from 'music' containing
# the returned images. The 'cover' column is updated to point to the most likely
# image returned. The search system for CDNow is quite complex but could be adapted
# to other sites. Google is simpler.
################################################################################ 
#
# background info :
#
# Cache : a simple cache is implemented to avoid pounding on CDNow for the same
# page over and over. If 'developing' is set to 1 then the cache is loaded / saved
# on startup, speeding the development of the fuzzy logic matching algorithms.


# TODO : clean up control-c and 'press q' hack.


use defsjukebox; # Get the settings entered by the user during configuration


# some settings you might care to adjust

my $imagemagick = 0;
my $oneshot = 0; # update only one cover then return
my $session_ID = "1951353473"; # CDNOW session ID. Might need updating periodically
my $developing = 0; # set to 1 to have the cache made bigger and load / save on entry / exit ( unless called from Jukebox )
my $maxMatchCDNow = 5;  # default
my $maxMatchGoogle = 20;
my $usecache = 1; # use cacheing to avoid pounding the cdnow servers. Set to 1 unless you suspect problems
my $recache = 0; # set to 1 and for each http request the cache entry is deleted and regenerated


# these get set via command line options.
my $skipjpgcreate = 0;
my $skipdatabaseupdate = 0;
my $verbose = 0;
my $usegoogle = 0;
my $deleteexisting = 0;
my $maxMatch = 0;   # the maximum number of artists & albums to try in cdnow.com for each Jukebox album
my $imgdir = '';
my $skip = 0;
my $fromscript = 0;
my $singleid = 0;
my $uploadfile = '';
my $renamename = '';
my $ripid = 0; # id in the ripqueue when called from jukebox
my $customsearch = '';
my $returnpath = 0;
my $cleanup = 0;


use Getopt::Long;

Getopt::Long::Configure("pass_through","no_ignore_case");

my $help=0;


GetOptions("help|h",\$help,
	   "maximages|m=i",\$maxMatch,
	   "nothing|n",\$skip,
	   "google|g",\$usegoogle,
	   "verbose|v",\$verbose,
	   "delete|d",\$deleteexisting,
	   "path|p=s",\$imgdir,
	   "jukebox|j",\$fromscript,
	   "id|i=i",\$singleid,
	   "file|f=s",\$uploadfile,
	   "rename|r=s",\$renamename,
	   "search|s=s",\$customsearch,
	   "returnpath",\$returnpath,
	   "cleanup",\$cleanup);

if ($help) {
    print <<__EOT__;
Usage:

    $0 {options}

    Walks the albums and attempts to find cover art for each entry. Search is done on CDNow or Google. Cover art is placed
    into a directory tree with each Album having an entry. Multiple images may be returned for each album since the search is
    rarely distinct. The most likely image is set in the database. The other returned images can be selected in the Albums
    page in the Jukebox. The program generates a lot of temporary files in /tmp/cover_art_cache. These can be useful if the
    script is run multiple times, so delete them when you\'re happy with the results.

Options:
  --h|help                Display this message.
  --g|google              Search Google instead of CDNow.
  --m|maximages={number}  Max number of returned images per album. Default is 5 for CDNow and 20 for Google.
  --v|verbose             Increase verbosity.
  --d|delete              Delete everything in the image directory for an album before lookup. Useful for development.
  --n|nothing             Do nothing - dry run.
  --p|path                Base path for images. Should be inside webserver docroot. Default is $SITE/cover_art/
  --j|jukebox             Used internally when being called from Jukebox. Command line args are different.
  --i|id                  Update just this Album id.
  --f|file                Set above Album id to this file. Does size conversion, copies it into tree. Updates DB.
  --r|rename              Name of file to be used when placed into tree for above option.
  --s|search              Use this string for google search
  --returnpath            Given an id ( -i ) returns the path based on Artist / Album name. Called from Jukebox
  --cleanup               For each Album with images inside 'cover_art' directory, cleans up redundant images

__EOT__
    exit 2;
}


umask(0);  # create files & dirs with wide open access.

use DBI;

use File::Basename; 
use File::Path; 
use LWP::UserAgent; 
use LWP::Simple; 
use HTTP::Request::Common qw(POST);
use HTML::TokeParser;
use Text::Metaphone;
use Compress::Zlib;

if($imagemagick) {
    require Image::Magick;
    import Image;
} else {
    require GD;
    import GD;
}

if ($developing) {
    require Data::Dumper;
    import Data;

    require Storable;
    import Storable qw(freeze thaw store retrieve);

    require Term::ReadKey;
    import Term qw(ReadMode ReadKey);

}


my $dbh=DBI->connect("dbi:mysql:database=$DBNAME",$DBUSER,$DBPASS) || die;


my $CDNOW = "http://www.cdnow.com";
my $cachedir = "/tmp/cover_art_cache/"; # take care if you change this. --cleanup does an rm -rf * on this
my $webcache = $cachedir."webcache/";
my $optimumimagesize = 170;
my $minimumimagesize = (($optimumimagesize * 50) / 100);
my $imageresizebelow = (($optimumimagesize * 80) / 100);
my $imageresizeabove = (($optimumimagesize * 120) / 100);

################################################################################ 
# see if we were called from Jukebox

if (int(@ARGV)==3) { # 3 parameters. We must be getting called from ripd
    $ripid = shift or die "need arg1";
    $singleid = shift or die "need arg2";
    $fromscript = 1;
    $usegoogle = 1;

    my $sth=$dbh->prepare( "select title from ripqueue where id=$ripid") || die;  # title contains the options & custom search string for a google search
    $sth->execute();
    ($customsearch) = $sth->fetchrow();

    $deleteexisting = 1 if($customsearch =~ s/^\s*--delete//);

    $usegoogle = 0 if($customsearch =~ s/^\s*--cdnow//);

    $customsearch =~ s/\s+$//g;    # strip trailing spaces
    $customsearch =~ s/^\s+//g;    # strip leading spaces

}


if($skip) {
    $skipjpgcreate = 1;
    $skipdatabaseupdate = 1;
}

if(!$maxMatch) {
    $maxMatch = $maxMatchGoogle if $usegoogle; # set this global
    $maxMatch = $maxMatchCDNow if !$usegoogle; # set this global
}

if (!$imgdir) {
    $imgdir = "$SITE/cover_art/";
}

mkdir($imgdir,0666);

if($cleanup) {
    system ("rm -r $cachedir") if ($cachedir =~ m%^/tmp/%);
} else {
    mkdir($cachedir,0666);
    mkdir($webcache,0666);
}

my $orgspc="a-zA-Z0-9";
$orgspc=~s/([^-a-zA-Z0-9])/\\$1/g;
my $maxChooseWords = 100000; # value when strings match exactly

my %artistlookupcache;
my @artistlookupkeys;
my $artistlookupmaxentries = ($developing) ? 1000 : 100;

my %googlelookupcache;
my @googlelookupkeys;
my $googlelookupmaxentries = ($developing) ? 1000 : 100;

my %urllookupcache;
my @urllookupkeys;
my $urllookupmaxentries = ($developing) ? 10000 : 500;

my $artistlookupcacheref;
my $artistlookupkeysref;
my $urllookupcacheref;
my $urllookupkeysref;
my $googlelookupcacheref;
my $googlelookupkeysref;


if(!$fromscript && $developing && -r $webcache."artistlookupcache.compressed") {
    $artistlookupcacheref = retrieve($webcache."artistlookupcache.compressed");
    $artistlookupkeysref = retrieve($webcache."artistlookupkeys");
    $urllookupcacheref = retrieve($webcache."urllookupcache.compressed");
    $urllookupkeysref = retrieve($webcache."urllookupkeys");
}
else
{
    $artistlookupcacheref = \%artistlookupcache;
    $artistlookupkeysref = \@artistlookupkeys;
    $urllookupcacheref = \%urllookupcache;
    $urllookupkeysref = \@urllookupkeys;
}

sub convertcache {
    foreach my $key (keys %$artistlookupcacheref )
    {
	print length $artistlookupcacheref->{$key}, " : ";
	my $compressdata = compress $artistlookupcacheref->{$key};
	print length $compressdata, "\n";

	$artistlookupcacheref->{$key} = $compressdata;
    }

    foreach my $key (keys %$urllookupcacheref )
    {
	print length $urllookupcacheref->{$key}, " : ";
	my $compressdata = compress $urllookupcacheref->{$key};
	print length $compressdata, "\n";

	$urllookupcacheref->{$key} = $compressdata;
    }

    store ($artistlookupcacheref, 'cache/artistlookupcache.compressed');
    store ($urllookupcacheref, 'cache/urllookupcache.compressed');

    exit(0);
}


sub catchcontrolc { # no longer used

    if ($developing) {
	$developing = 0; # don't try and do it again
	dumpdatatodisk();
    }
    die;
}

sub dumpdatatodisk { # save my huge hash tables to make development faster ( cdnow lookups over ISDN are killing me )

    store ($artistlookupcacheref, $webcache."artistlookupcache.compressed");
    store ($artistlookupkeysref, $webcache."artistlookupkeys");
    store ($urllookupcacheref, $webcache."urllookupcache.compressed");
    store ($urllookupkeysref, $webcache."urllookupkeys");

    print "dumped from artist hash : ".(scalar @$artistlookupkeysref)." entries\n" if $verbose >= 1;
    print "dumped from url hash : ".(scalar @$urllookupkeysref)." entries\n" if $verbose >= 1;
}

sub fetchsetting {
    my $res=$dbh->prepare("select value from settings where name = ?");
    $res->execute($_[0]);
    my $ret;
    ($ret)=$res->fetchrow_array();
    return $ret;
}

sub stripthe {
    my ($str)=@_;
    $str=~s/(\.,)?(the)$//i;
    return $str;
}

sub strippunct {
    my ($str)=@_;

    $str =~ s/[\'\`]//g; # convert apostrophe to nothing 
    $str =~ s/[:;\-\*_\(\)!\[\]\"\'\?\\\/]/ /g;    # covert junk to spaces
    $str =~ s/\s+$//g;    # strip trailing spaces
    $str =~ s/^\s+//g;    # strip leading spaces
    $str =~ s/[,]/ /g;   # change ,
    return $str;
}

sub reducewords {
    my ($str)=@_;
    $str =~ s/&/and/i; # if we have an & character, convert it to and ( & may confuse the search engine )
    $str =~ s/(\b(disc|disk))\s?\d?/ /i; # remove disc numbers from multi-cd albums
    $str =~ s/\bvol\s?\d?/ /i; # and again
    $str =~ s/\bcd\s?\d?/ /i; # and again
    $str =~ s/\setc[\.,]?/ /i; # strip 'etc.'
    $str =~ s/soundtrack$/ /i; # strip soundtrack

    $str =~ s/\s+$//g;    # strip trailing spaces
    $str =~ s/^\s+//g;    # strip leading spaces
    return $str;
}

sub stripforcompare {
    my ($words1)=@_;

    $words1 =~ tr/A-Z/a-z/;
    $words1 =~ s/(\bthe\b)//g; # remove the from most places
    $words1 =~ s/the$//;  # and the end
    $words1 =~ s/\s+/ /g;    # strip multiple spaces
    $words1 =~ s/\s+$//g;    # strip trailing spaces
    $words1 =~ s/^\s+//g;    # strip leading spaces

    return $words1;
}

sub unSpecial {
    my ($str)=@_;
    $str=~s/[^$orgspc]+/_/g;
    return $str;
}

sub urlToFileName {
    my ($str)=@_;
    $str=~s%^http://%%;
    return $str;
}


sub constructPath {
    my $artist = shift or die "need artist";
    my $album = shift or die "need album"; 

    my $dirAlbum = "$imgdir".unSpecial($artist);
    mkdir( $dirAlbum,0666 ); 
    $dirAlbum .= "/".unSpecial($album)."/";
    mkdir( $dirAlbum,0666 ); 

    return $dirAlbum;
}

sub constructRelativePath {
    my $artist = shift or die "need artist";
    my $album = shift or die "need album"; 

    my $dirAlbum = unSpecial($artist)."/".unSpecial($album)."/";

    return $dirAlbum;
}

################################################################################ 

my $ua = new LWP::UserAgent or die $!; 
$ua->timeout( 20 ); 
my $hdrs = new HTTP::Headers( Accept => '*/*', 'User-Agent' => 'gjukebox' ); 
my $gHTTPFailures  = 0;
my $gMaxHTTPFailures = 10;

################################################################################ 

if(!$oneshot && !$fromscript && $developing) {
    open(TTY, "</dev/tty");
    Term::ReadKey::ReadMode (3);
}

my $sth;


if($singleid) {
    $sth=$dbh->prepare( "select name,artist,id,cover from programs where id=$singleid") || die; 
} elsif($cleanup)  { # delete unwanted files from directories
    $sth=$dbh->prepare( "select name,artist,id,cover from programs where cover is not null order by artist") || die; 
} else { # look for albums with missing covers 
    $sth=$dbh->prepare( "select name,artist,id,cover from programs where cover is null order by artist") || die; 
}

$sth->execute();
while (my($album,$artist,$id,$cover) = $sth->fetchrow() ) {

    $artist = "Various" if(!$artist); # use capital V to be consistent with php code.

    my $images = -1;
    my $firstpath;

    if(!$oneshot && !$fromscript  && $developing) {
	my $key = Term::ReadKey::ReadKey (-1, *TTY);  # quit if the user presses 'q'. Better than Ctl-C because my data dumper gets called
	if ($key && $key eq 'q') {
	    last;
	}
    }

    if($returnpath && $singleid) {
	print constructPath($artist,$album);  # print the cover art path for this Album record

    } elsif($uploadfile && $renamename && $singleid) {  # we've been passed a file to place into the tree

	$firstpath = prepareExistingImage ($artist, $album, $uploadfile, $renamename);
	$images = 1 if $firstpath;

    } elsif($cleanup) {
	cleanupFilesForAlbum($artist, $album, $cover);
    } else {
	my $editalbum = reducewords(strippunct($album));
	my $editartist = reducewords(strippunct( stripthe($artist) ));

	$editalbum = stripforcompare($editalbum);
	$editartist = stripforcompare($editartist);

	my $matchon = "artist";

	if($editartist =~ /^various/i) {
	    print "Artist is various. Matching on album name\n" if $verbose >= 1;
	    $matchon = "album";
	}

	print ("====>$editartist : $editalbum<====") if $verbose >= 1;

	if($customsearch && $usegoogle) { # custom search string from the form in programs.php
	    $matchon = 'custom';
	    $editalbum = $customsearch;
	}

	# do the search and return the first ( most likely ? ) image found

	($images, $firstpath) = findcover($editartist, $editalbum, $artist, $album, $matchon);

    }

    if($images >= 0) {
	my $update =$dbh->prepare( "update programs set cover=? where id=?") || die;
	$update->execute($firstpath,$id) if!$skipdatabaseupdate;
    }

    last if( $oneshot) # for debugging

}

if(!$oneshot && !$fromscript && $developing) {
    Term::ReadKey::ReadMode (0);
}

dumpdatatodisk() if $developing && !$fromscript;

$dbh->disconnect;

print "\n";

################################################################################ 

sub findcover 
{ 
    my ($editartist, $editalbum, $artist, $album, $matchon ) = @_; 
    my $images;
    my $firstpath;

    ($images, $firstpath) = find_disccover( $editartist, $editalbum, $artist, $album, $matchon ) ;

    return $images, $firstpath;

} 


################################################################################ 

sub find_disccover 
{ 
    my $images;
    my $firstpath;

    if(!$usegoogle) {
	($images, $firstpath) = &find_disccover__cdnow; 
    } else {
	($images, $firstpath) = &find_disccover__google; 
    }

    return $images, $firstpath; 
} 

################################################################################ 

sub find_disccover__cdnow 
{ 
    my ($editartist, $editalbum, $artist, $album, $matchon ) = @_; 

    my $dirAlbum = constructPath($artist, $album);

    system("rm $dirAlbum*") if $deleteexisting; # delete all the existing pictures ?

    my ($images, $firstname) = cdnow( $editartist, $editalbum, $dirAlbum, $matchon);

    if($firstname) {
	$firstname = constructRelativePath($artist, $album).$firstname;   # firstpath is really just the name. Make it the relative path here
    }

    return $images, $firstname; 

} 

################################################################################ 

sub find_disccover__google 
{ 
    my ($editartist, $editalbum, $artist, $album, $matchon ) = @_; 

    my $rc = 0; 

    my $dirAlbum = constructPath($artist, $album);

    system("rm $dirAlbum*") if $deleteexisting;

    my ($images, $firstname) = google( $editartist, $editalbum, $dirAlbum, $matchon);

    if($firstname) {
	$firstname = constructRelativePath($artist, $album).$firstname;   # firstpath is really just the name. Make it the relative path here
    }

    return $images, $firstname; 

}


################################################################################ 

sub google {
    my $findArtist = shift or die "need artist and album"; 
    my $findAlbum = shift or die "need album"; 
    my $dir = shift or die "need directory"; 
    my $matchon = shift or die "need matchon"; 

    my $numimages = 0;
    my $firstpath;
    my $path;
    my $name;
    my $picture;

    print "\n" if $verbose >= 1;

#    open( LOG,  ">/tmp/google.log" ) or die; 

    my $iteration = 0;
    my $maxiterations = ($matchon eq "custom") ? 1 : 4;

    while ($numimages < $maxMatch && $iteration < $maxiterations) {

        my $search;

	if(($matchon eq "album") || ($matchon eq "custom")) {
	    $search = "$findAlbum";
	} else {
	    $search = "$findArtist $findAlbum" if ($iteration == 0);
	    $search = "$findAlbum $findArtist" if ($iteration == 1);
	    $search = "$findAlbum" if ($iteration == 2);
	    $search = "$findArtist" if ($iteration == 3);
	}

	$search =~ s/(\b([a-zA-Z]{0,2})\b)//g if ($iteration >= 0); # strip small words
	$search =~ s/\.//g if ($iteration >= 0); # strip . 
	$search =~ s/\b(and|the)\b//g if ($iteration >= 0);
	$search =~ s/\s+/ /g;    # strip multiple spaces
	$search =~ s/\s+$//g;    # strip trailing spaces
	$search =~ s/^\s+//g;    # strip leading spaces

	print "---->iteration $iteration. Google search for \"$search\"..." if $verbose >= 1;
#	print LOG "---->iteration $iteration. Google search for \"$search\"...";

	my $searchurl = "http://images.google.com/images?q=$search&num=50";

	my $googlepage = cached_get_from_url($searchurl); 

	return -1 if !$googlepage;

	my @jpeglinks;
	my $numlinks;

	my ($p) = HTML::TokeParser->new(\$googlepage);

	# go through looking for links. If we find them, see if they point to a discography. If it does
	# check that it's absolute, and bolt the cdnow base URL onto it if not. Then push it onto an array

	my $numLinks = 0;

	$p->{textify} = { };
	
	while (my $token = $p->get_tag("a")) {
	    my $url = $token->[1]{href} || "-";
	    my $text = $p->get_trimmed_text("/a");
	    if($url =~ /imgurl/i) {

#<a href=/imgres?imgurl=www.classicalguitarstudio.com/bach213.jpg&imgrefurl=http://www.classicalguitarstudio.com/bach213.htm&h=316&w=230&prev=/images%3Fq%3Dbach%2Bj.s.%2Bcello%2B%26num%3D20%26hl%3Den%26sa%3DG>

		$url =~ s/(.*?)=//;
		$url =~ m/(.*?)&/;
		$url = $1;

		$jpeglinks[$numLinks][0] = $url;
		$jpeglinks[$numLinks][1] = $text;

		$numLinks += 1;
	
	    }
	}
	
	my $i = 0;
	print "\n" if $verbose >= 1;
	my $gotimageok;

	while(($i < $numLinks) && ($numimages < $maxMatch)) {
	    
	    $gotimageok = 0;
	    my $imgurl = $jpeglinks[$i][0];

	    if($imgurl =~ m/^http/) {
	    } else {  # why the fuck can I not work out how to say 'doesn't match' in a RegExp ? I give up
		$imgurl = "http://".$imgurl;
	    }

	    print "  trying $imgurl..." if $verbose >= 1;
#	    print LOG "  trying $imgurl...";

	    ($picture, $name) = extract_image($imgurl);

	    if($picture) {

		my $path = store_image($picture, $dir, $name, $i);

		if($path) {
		    print " ok\n" if $verbose >= 1;
		    $numimages += 1;
		    $firstpath = $path if !$firstpath;
		    $gotimageok = 1;
		}
	    }
	    print "\n" if (!$gotimageok && $verbose >= 1);

	    print "got $numimages%\n" if ($fromscript);
#	    print LOG "got $numimages%\n" if ($fromscript);

	    $i++;

	}

	 $iteration++ if($numimages < $maxMatch);

    }

    return($numimages, $firstpath);

} 


################################################################################ 

sub cdnow 
{

    my $findArtist = shift or die "need artist"; 
    my $findAlbum = shift or die "need album"; 
    my $dir = shift or die "need directory"; 
    my $matchon = shift or die "need matchon"; 
    my $images;
    my $imagenum = 0;
    my $firstname;
    my $name;
    my @artistlinks;

    print "..search.." if $verbose >= 1;
    my $discography = cached_cdnow_artistlookup($findArtist, $findAlbum, $matchon); 
    print"\n" if $verbose >= 1;

    $findArtist = $findAlbum if $matchon eq "album"; # make them the same to fool later matching code

    return -1 if(!$discography);

    if( $discography =~ /Next Page/i) {
	print "Warning -----------------> multiple pages in discography\n" if $verbose >= 1;
    }

    if ( $discography =~ /more than one artist matched/i ||
	 $discography =~ /following results match closest/i) 
    { 
	# CDnow presents a 'select an atist page' if the match is ambigous.

	my ($p) = HTML::TokeParser->new(\$discography);

	# go through looking for links. If we find them, see if they point to a discography. If it does
	# check that it's absolute, and bolt the cdnow base URL onto it if not. Then push it onto an array

	my $numLinks = 0;

	$p->{textify} = { };
	
	while (my $token = $p->get_tag("a")) {
	    my $url = $token->[1]{href} || "-";
	    my $text = $p->get_trimmed_text("/a");
	    if($url =~ /discography\.html\/ArtistID/i) {
		if(!($url =~ /^http/)) {
		    $url = $CDNOW.$url;
		}
		$text = reducewords(strippunct($text));
		$text = stripforcompare($text);

		if ($text) {
		    $artistlinks[$numLinks][0] = $url;
		    $artistlinks[$numLinks][1] = $text;
	
		    $artistlinks[$numLinks][2] = comparewords($text, $findArtist); # how close does this match the name we want ?
		    
		    $numLinks += 1;
		}
	    }
	}

	my $i;
	my $exactmatch = 0;
	$discography = '';
        my $onedisco = '';

	@artistlinks = sort my_compare @artistlinks;

	# go through list of matching artists and get the discography for each one

	for($i=0; $i<$numLinks; $i++) {

	    print"  Artist close match : $artistlinks[$i][1]...fetching discography..." if $verbose >= 1;
	    $onedisco = cached_get_from_url($artistlinks[$i][0]); 
	    print"\n" if $verbose >= 1;
	    return -1 if !$onedisco;

	    $artistlinks[$i][3] = $onedisco;

	    if($i >= 5 && $numLinks > 5) {
		print "  skipping ".($numLinks - $i)." other artists where match is less than ".$artistlinks[$i][2]."\n" if $verbose >= 1;
		last;
	    }
	}			
    
	if ($i>0) {
	    ($images, $name) = cdnow_parse_discography($findAlbum, $dir, $imagenum, \@artistlinks, $i);
	}

	$firstname = $name if !$firstname;
	return -1 if ($gHTTPFailures > $gMaxHTTPFailures);

	$imagenum += $images;

      
    }

    else {

	$artistlinks[0][0] = "dummy url";
	$artistlinks[0][1] = $findArtist;
	$artistlinks[0][2] = $maxChooseWords;
	$artistlinks[0][3] = $discography;

	print"  Artist exact match : $findArtist\n" if $verbose >= 1;
	($images, $name) = cdnow_parse_discography($findAlbum, $dir, $imagenum, \@artistlinks, 1);
	$firstname = $name if !$firstname;

	return -1 if $images == -1;
	$imagenum += $images;
    }

    return $imagenum, $firstname; 

}

############################################################################### 


sub cdnow_parse_discography {

    my $findAlbum = shift or die "need artist and album"; 
    my $dir = shift or die "need directory"; 
    my $beginimagenum = shift; 
    my $artistlinks = shift or die "need albumlinks";
    my $numartists = shift or die "need numalbums";
 
    my @albumlinks;
    my $albumPage;
    my $totalalbums = 0;

    for(my $artist = 0; $artist < $numartists; $artist++) {

	# for each artist we grab & parse the discography getting the list of album links & corresponding album names
	# albums extracted are appended onto albumlinks start at index totalalbum. Return value is the number appended
	# element 3 is the discography HTML

	my ($numalbums) = extractAlbums( $findAlbum, $$artistlinks[$artist][3], $$artistlinks[$artist][2], \@albumlinks, $totalalbums); 
	$totalalbums += $numalbums;
	
    }

    # now the magic sort on element[i][2]. The bigger the value, the closer both artist and album match

    @albumlinks = sort my_compare @albumlinks;

    if($verbose >=1) {
	for(my $i = 0; $i < $totalalbums; $i++) {
	    print "    sorted discography : $albumlinks[$i][1]\n" if $verbose >= 1;
	}
    }

    my $finalurl;
    my $picture;


    my $i = 0;
    my $numimages = $beginimagenum;
    my $exact = 0;
    my $name;
    my $firstpath;

    while($albumlinks[$i][0] && $numimages < $maxMatch) {

	print "    reading page for \'$albumlinks[$i][1]\'..." if $verbose >= 1;
	$albumPage = cached_get_from_url($albumlinks[$i][0]);
	print "\n" if $verbose >= 1;

	if(!$albumPage) {
	    return(-1) if ($gHTTPFailures > $gMaxHTTPFailures)
	}

	print "    checking \'$albumlinks[$i][1]\' for cover art..." if $verbose >= 1;

    #<TD ALIGN=CENTER width=200><img src="/graphics/spacer.gif" width=200 height=1><br>\<img src="http://gs.cdnow.com/graphics/COVERART/local/L/56/88/00015688.jpg"; ALT="Co\ver Image" width=170 height=170 border=2> 

	( my $imgurl ) = $albumPage =~ m#"(http://[^"]*/COVERART/[^"]*)"#; 

    # cdnow returns either http://gs.cdnow.com/... or /cgi-bin/... 

	if($imgurl) {
	
	    $_ = $imgurl;

	    if(/^http/) {
		$finalurl = $imgurl;
	    } else {
		$finalurl = $CDNOW.$imgurl;
	    }

	    ($picture, $name) = extract_image($finalurl);

	    if($picture) {
#		print "\n" if $verbose >= 1;

		if(comparewords($albumlinks[$i][1], $findAlbum) >= $maxChooseWords) {
		    print "    Album Exact match : $albumlinks[$i][1]" if $verbose >= 1;
		    $exact = 1;
		} else {
#		    print "    Album Close match : $albumlinks[$i][1]";
		}

		my $path = store_image($picture, $dir, $name, $i + $beginimagenum);

		if($path) {
		    $numimages += 1;
		    $firstpath = $path if !$firstpath;
		}
	    }
	}
       	if (( !$imgurl || !$picture) ) {
	    print "none found\n" if $verbose>=1;
	} else {
	    print "\n" if $verbose>=1;
	}

	$i++;

	last if($exact && ($picture)); # if the album name matches exactly then don't bother getting any more
    }

    return $numimages, $firstpath;
} 

################################################################################ 
# sort function called in chooseAlbum

sub album_compare_old
{
    my $aalbum;
    my $balbum;

    $aalbum = abs(adist($_, $a->[1]));#element 1 contains the album text. Compare it to the global gAlbumName
    $balbum = abs(adist($_, $b->[1]));

    my $alength = 0; # abs(length($gAlbumName) - length($a->[1])) / 3;
    my $blength = 0; # abs(length($gAlbumName) - length($b->[1])) / 3;

    return(($aalbum + $alength) <=> ($balbum + $blength));

}

# compare one string to another, returning a value that is larger depending on how close they are. Value could be negative.
# used in the sort function to rank how close an album / artist is to the one we want.

sub comparewords {
    my $words1 = shift or die "need words1";
    my $words2 = shift or die "need words2";

    #print "comparing $words1 : $words2\n";

    my $longer;
    my $shorter;
    my $longerwordlengths;
    my $shorterwordlengths;
    my $longlen;
    my $shortlen;
    my @metawords1;
    my @metawords2;
    my @lengths1;
    my @lengths2;

    if($words1 eq $words2) {
#	print "comparing $words1 : $words2 --- closeness is $maxChooseWords\n";
	return $maxChooseWords ;
    }

    push @metawords1, split / /, $words1;  # break it into an array

    for(my $i = 0; $i < scalar @metawords1; $i++) { # convert each word to phonemes to bypass spelling problems. Not all words conver, so
	$lengths1[$i] = length( $metawords1[$i]);
	$lengths1[$i] = 0 if(!$lengths1[$i]);
	my $junk = Metaphone( $metawords1[$i] );    # leave intact those
	$metawords1[$i] = $junk if($junk);
    }

    push @metawords2, split / /, $words2;

    for(my $i = 0; $i < scalar @metawords2; $i++) {
	$lengths2[$i] = length( $metawords2[$i]);
	$lengths2[$i] = 0 if(!$lengths2[$i]);
	my $junk = Metaphone( $metawords2[$i] );
	$metawords2[$i] = $junk if($junk);
    }

    if(scalar @metawords2 > scalar @metawords1) {   # work out which list is longer
	$longer = \@metawords2;
	$longerwordlengths = \@lengths2;
	$longlen = scalar @metawords2;
	$shorter = \@metawords1;
	$shorterwordlengths = \@lengths1;
	$shortlen = scalar @metawords1;
    } else {
	$longer = \@metawords1;
	$longerwordlengths = \@lengths1;
	$longlen = scalar @metawords1;
	$shorter = \@metawords2;
	$shorterwordlengths = \@lengths2;
	$shortlen = scalar @metawords2;
    }

    my $closeness = 0;
    my $found = 0;
    my $distance;

    for(my $wordshort = 0; $wordshort < $shortlen; $wordshort++) {
	for(my $wordlong = 0; $wordlong < $longlen; $wordlong++) {
	    if ($shorter->[$wordshort] eq $longer->[$wordlong]) {        # words match
		if ($wordshort == $wordlong) {  # same position
		    $closeness += (10 * $longerwordlengths->[$wordlong]);
		} else {
		    $distance = (10 - abs($wordshort - $wordlong));
		    $closeness += ((($distance < 5) ? 1 : $distance) * $longerwordlengths->[$wordlong]); # factor in how far apart the words are
		}
		$found = 1;
	    }
	}
	$closeness -= $shorterwordlengths->[$wordshort] if !$found;

    }
    $closeness -= ($longlen - $shortlen);

#    print "comparing $words1 : $words2 --- closeness is $closeness\n";

    return $closeness;
}

sub my_compare
{
    return ($b->[2] <=> $a->[2] ); # order of comparison is inverted because comparewords returns bigger for better match
}

################################################################################ 
sub extractAlbums 
{ 
    my ( $findAlbum, $discogHtml, $artistcloseness, $returnarray, $arraysize ) = @_; 
    
    my ($p) = HTML::TokeParser->new(\$discogHtml);

    # go through looking for links. If we find them, see if they point to the album ( look at text of the link ). If it does
    # check that it's absolute, and bolt the cdnow base URL onto it if not. Then push it onto an array

    $p->{textify} = { };

    my $numLinks = 0;

    while (my $token = $p->get_tag("a")) {
	my $url = $token->[1]{href} || "-";
	my $text = $p->get_trimmed_text("/a");
	$_ = $url;
	if(/album.html/i && $text) {
	    $text = reducewords(strippunct($text)); #reduce this album name the same way we did our one from DB
	    $text =~ tr/A-Z/a-z/;

	    $_ = $url;
	    if(!/^http/) {
		$url = $CDNOW.$url
	    }
	    $text = reducewords(strippunct($text));
	    $text = stripforcompare($text);
	    if ($text) {
		$$returnarray[$arraysize][0] = $url;
		$$returnarray[$arraysize][1] = $text;
		$$returnarray[$arraysize][2] = $artistcloseness + comparewords($text, $findAlbum); # how close is this match
		$numLinks += 1;
		$arraysize += 1;
	    }
	}
    }

    return ($numLinks); 
} 

################################################################################ 

sub extract_image
{ 
    my ( $imgurl ) = @_;
    my $tempimage = "tempimage";
    my ($img,  $width, $height, $format);

    my ($path, $name) = ($imgurl =~ m|^(.*?)/?([^/]+/*)$|);

    return unless $name; 

    if($imagemagick) {
	return unless $name =~ /\.(jpe?g|png|gif)$/i; # filter out the wrong exts.
	$tempimage.=".".$1; # Set extension to tempname
    } else {
	return unless $name =~ /\.(jpe?g|png)$/i; # filter out the wrong exts.
	$tempimage.=".".$1; # Set extension to tempname
    }

    print " getting picture : $name.." if (!$usegoogle && $verbose >= 1); # don't print for Google - already printed in calling func.

# transfer URL to a valid filename. Slight chance of problems with two URLs pointing to one file. cdnow seems OK.

    my $cache = urlToFileName($imgurl); 

    my $picture;

    if(! -r $cachedir.$cache) { # we didn't already cache this file
    
	my $req = new HTTP::Request( GET => $imgurl, $hdrs ); 
	my $resp; 
	my $attempt = 0;

	do {
	    $resp = $ua->request( $req ); 

	    if( !$resp->is_success ) {
		print "\nHTTP error  ".$resp->status_line."\n" if $verbose >= 1;
	    }

	} while ( ++$attempt < 3 && !$resp->is_success && !( $resp->status_line =~ m/^4\d\d/));

	if(!$resp->is_success) {
	    print "Too many retries. Exiting.\n" if($attempt >= 3 && $verbose >= 1);
	    $gHTTPFailures += 1 if(!$usegoogle);  # don't keep pounding cdnow if too many errors.
	    return(0);
	}

	open( IMG,  ">$cachedir$tempimage" ) or die $!; 
	print IMG $resp->content; 
	close( IMG ); 
    
	$picture = $resp->content;

	if($imagemagick) {
	    # get info about image using imagemagick [width,height,format] 
	    #############################################################
    
	    $img = Image::Magick->new(); 
	    $img->Read( filename =>  $cachedir.$tempimage ); 
	    ( $width, $height, $format ) = $img->Get( qw( width height format ) ); 

	    # reject malformed images 
	    return 0 unless $format; 

	    # reject [approximately] non-square images 
	    my $imageRatio = $width / $height; 
	    my $error = abs( $imageRatio - 1 ); 

	    if ($error > 0.2) {
		print "rejected non square image" if $verbose >= 1;
		return 0;
	    }

	    if( $width < $minimumimagesize || $height < $minimumimagesize) {
		print "rejected small image" if $verbose >= 1;
		return 0;
	    }

	    # now cache it.
	    mkpath ( dirname ( $cachedir.$cache ),0,0666);

	    if( $width < $imageresizebelow || $width > $imageresizeabove ||
			    $height < $imageresizebelow || $height > $imageresizeabove ) {

		print "..resizing image" if( $verbose >= 1);

		$img->Resize( width => $optimumimagesize, height => ($optimumimagesize * ($height / $width)) );
		$img->Write( filename =>  $cachedir.$cache ); 
    
		$picture = `cat $cachedir$cache`;  # now just read back whole file.
	    }
    
	} else {

	    # No Image magick so we'll use GD instead
	    ######################################################

	    if ($tempimage=~/jpe?g$/i) {
		$img = GD::Image->newFromJpeg($cachedir.$tempimage);
	    } elsif ($tempimage=~/png$/i) {
		$img = GD::Image->newFromPNG($cachedir.$tempimage);
	    }

	# reject malformed images 

	    return 0 unless $img; 

	    ($width,$height) = $img->getBounds();

	# reject [approximately] non-square images 
	    my $imageRatio = $width / $height; 
	    my $error = abs( $imageRatio - 1 ); 

	    if ($error > 0.2) {
		print "rejected non square image" if $verbose >= 1;
		return 0;
	    }

	    if( $width < $minimumimagesize || $height < $minimumimagesize) {
		print "rejected small image" if $verbose >= 1;
		return 0;
	    }

	    # now cache it.
	    mkpath ( dirname ( $cachedir.$cache ),0,0666);

	    if( $width < $imageresizebelow || $width > $imageresizeabove ||
			    $height < $imageresizebelow || $height > $imageresizeabove ) {

		print "..resizing image" if( $verbose >= 1);

		my $newImg=new GD::Image($optimumimagesize,$optimumimagesize * ($height / $width));
		$newImg->copyResized($img,0,0,0,0,
				     $optimumimagesize,$optimumimagesize*($height/$width),
				     $width,$height);

		open(OUT,">$cachedir$cache") || die;
		if ($cache=~/jpe?g$/i) {
		    print OUT $newImg->jpeg;
		} elsif ($cache=~/png$/i) {
		    print OUT $newImg->png;
		}
		close OUT;

		$picture = `cat $cachedir$cache`;  # just read whole file.

	    } else {
		open( IMG,  ">$cachedir$cache" ) or die $!; 
		print IMG $picture; 
		close( IMG ); 
	    }

    	}

    } else {
	print " from cache" if $verbose >= 1;
	$picture = `cat $cachedir$cache`;  # just read whole file.
    }


    return $picture, $name; 
} 

################################################################################ 
# take a previously manipulated image and place it into the tree, making the 
# name unique if necessary

sub store_image
{
    my ( $image, $dir, $name, $numimages ) = @_; 
    return 0 unless $image; 
    return 0 if $skipjpgcreate;

    my $file = "$dir/$name"; 

    if (-r $file) { # file exists ? save it under a new name

	if($usegoogle) {
	    print "  renaming." if $verbose >= 1;
	    $file =~ s/\./$numimages\./;
	    $name =~ s/\./$numimages\./;
	} else {
	    print "  already have it." if $verbose >= 1;
	    return 0;
	}

    }
    print "  writing $name" if $verbose >= 1;

    open( IMG, ">$file" ) or die "$file: $!"; 
    print IMG $image; 
    close( IMG ); 

    return $name; 
} 

################################################################################ 

sub prepareExistingImage {
    my $artist = shift or die "need artist"; 
    my $album = shift or die "need album"; 
    my $infilename = shift or die "need src filename";
    my $renamename = shift or die "need target filename";

    my ($img,  $width, $height, $format, $name);


    my $i = 1;

    my $dirAlbum = constructPath($artist, $album);
    $name = $dirAlbum.basename($renamename);

    while (-r $name && $i < 1000) { # file exists ? save it under a new name

	print "  renaming.\n" if $verbose >= 1;
	$name =~ s/[0-9]*\./$i\./;
	$i++;
    }

    return 0 if($i >= 1000);

    
    if($imagemagick) {
    
	$img = Image::Magick->new(); 
	$img->Read( filename =>  $infilename ); 
	( $width, $height, $format ) = $img->Get( qw( width height format ) ); 

    # reject malformed images 
	return 0 unless $format; 

	if( $width < $minimumimagesize || $height < $minimumimagesize) {
	    print "rejected small image" if $verbose >= 1;
	    return 0;
	}

	if( $width < $imageresizebelow || $width > $imageresizeabove ||
			$height < $imageresizebelow || $height > $imageresizeabove ) {

	    print "..resizing image\n" if( $verbose >= 1);

	    $img->Resize( width => $optimumimagesize, height => ($optimumimagesize * ($height / $width)) );

	} 
	print "..saving to $name\n" if $verbose >= 1;

	$img->Write( filename =>  $name ); 

    
    } else {
    
	 if ($infilename=~/jpe?g$/i) {
	    $img = GD::Image->newFromJpeg($infilename);
	} elsif ($infilename=~/png$/i) {
	    $img = GD::Image->newFromPNG($infilename);
	}

    # reject malformed images 
	return 0 unless $img; 

	($width,$height) = $img->getBounds();

	if( $width < $minimumimagesize || $height < $minimumimagesize) {
	    print "rejected small image" if $verbose >= 1;
	    return 0;
	}

    	if( $width < $imageresizebelow || $width > $imageresizeabove ||
			$height < $imageresizebelow || $height > $imageresizeabove ) {

	    print "..resizing image\n" if( $verbose >= 1);

	    my $newImg=new GD::Image($optimumimagesize,$optimumimagesize * ($height / $width));
	    $newImg->copyResized($img,0,0,0,0,
				 $optimumimagesize,$optimumimagesize*($height/$width),
				 $width,$height);

	    $img=$newImg;

	} 
	print "..saving to $name\n" if $verbose >= 1;

	open(OUT,">$name") || die;
	if ($name=~/jpe?g$/i) {
	    print OUT $img->jpeg;
	} elsif ($name=~/png$/i) {
	    print OUT $img->png;
	}
	close OUT;
    }

    return constructRelativePath($artist, $album).basename($name);

}


################################################################################ 
# do an artist lookup on cdnow.com. Also uses a cache mechanism to stop
# cdnow getting unhappy about the pounding on their servers

sub cached_cdnow_artistlookup {
    my $findArtist = shift or die "need artist"; 
    my $findAlbum = shift or die "need album"; 
    my $matchon = shift or die "need matchon"; 
    my $cacheName;
    my $req;

    # cd now main page -- search on artist 
    # /SID=1908667405

    my $dst = "$CDNOW/cgi-bin/mserver/SID=$session_ID/pagename=/RP/CDN/FIND/popsearch.html/clickID=tn_srch_txt"; 

    # scalar keys %artistlookupcache

    if($matchon eq "album") { # match on album - usually for compilations & soundtracks
	$cacheName = $findAlbum."\@album2"; # change the key used for the cache to ensure different results for album & artist cache entries
        $req = POST $dst, [ index => 't', string => $findAlbum ]; 
    } else {
	$cacheName = $findArtist;
	$req = POST $dst, [ index => 'a', string => $findArtist ]; 
    }

    if($usecache && $recache)   { # flush this entry from the cache
	delete ($$artistlookupcacheref {$cacheName});
    }

    if($usecache && $artistlookupcacheref->{$cacheName})   {
	print "  from cache" if $verbose >= 1;
	return uncompress $artistlookupcacheref->{$cacheName};
    }

    my $resp; 

    my $attempt = 0;

    do {
	$resp = $ua->request( $req ); 

	if( !$resp->is_success ) {
	    print "\nHTTP error  ".$resp->status_line."\n" if $verbose >= 1;
	    print Dumper( $resp ), "\n" if $verbose >=3; 
	}

    } while ( $attempt++ < 3 && !$resp->is_success && !($resp->status_line =~ m/^4\d\d/));

    if(!$resp->is_success) {
	print "Too many retries.\n"  if($attempt >= 3 && $verbose >= 1);
	$gHTTPFailures += 1 if(!$usegoogle);  # don't keep pounding cdnow if too many errors.
	return(0);
    }

    if($usecache) {
	if(scalar @$artistlookupkeysref > $artistlookupmaxentries) # exceeded number of cache entries ?
	{
	    delete( $$artistlookupcacheref { shift @$artistlookupkeysref } );  # delete the first key/value pair
	}
	push @$artistlookupkeysref, $cacheName;

	$artistlookupcacheref->{$cacheName} = compress $resp->content;
    }

    return $resp->content;

}


################################################################################ 
# do an artist lookup on cdnow.com. Also uses a cache mechanism to stop
# cdnow getting unhappy about the pounding on their servers

sub cached_get_from_url {

    my $url = shift or die "need url"; 

    if($usecache && $recache)   { # flush this entry from the cache
	delete ($$urllookupcacheref {$url});
    }

    if($usecache && $urllookupcacheref->{$url})   {
	print "  from cache" if $verbose >= 1;
	return uncompress $urllookupcacheref->{$url};
    }

    my $req = new HTTP::Request(GET => $url, $hdrs); 
    my $resp; 

    my $attempt = 0;

    do {
	$resp = $ua->request( $req ); 

	if( !$resp->is_success ) {
	    print "\nHTTP error  ".$resp->status_line."\n" if $verbose >= 1;
	    print Dumper( $resp ), "\n" if ($verbose >=3 && $verbose >= 1); 
	}

    } while ( $attempt++ < 3 && !$resp->is_success && !($resp->status_line =~ m/^4\d\d/));

    if(!$resp->is_success) {
	print "Too many retries.\n" if($attempt >= 3);
	$gHTTPFailures += 1 if(!$usegoogle);  # don't keep pounding cdnow if too many errors.

	return(0);
    }

    if($usecache) {
	if(scalar @$urllookupkeysref > $urllookupmaxentries) # exceeded number of cache entries ?
	{
	    delete( $$urllookupcacheref { shift @$urllookupkeysref } );  # delete the first key/value pair
	}
	push @$urllookupkeysref, $url;

	$urllookupcacheref->{$url} = compress $resp->content;
    }

    return $resp->content;
}


################################################################################ 
# delete all files for an album except the one selected in the DB

sub cleanupFilesForAlbum
{ 
    my ( $artist, $album, $selectedfile ) = @_; 

    my $directory = constructRelativePath($artist, $album)."temp"; # the temp bit is just to make dirname return the path OK.

    return if(!(dirname ( $directory ) eq dirname ( $selectedfile )));
    
    $directory = constructPath($artist, $album);

    opendir(COVERDIR, $directory) or return;
    my @files = readdir COVERDIR;
    closedir COVERDIR;
    
    foreach my $file (@files) {
	if($file ne '.' && $file ne '..' && $file ne basename($selectedfile)) {
	     print "deleting $directory$file\n";
	     system("rm $directory$file");
	}
    }
}
